#' Decision Tree
#'
#' This function recursively divides the sets of observations
#' into two subsets and create a decision tree. \cr
#' Japanese help file: \link[lazy.tree]{DecTree_JPH}
#'
#' @param y The criterion variable (as vector)
#' @param X The regressor matrix (as data frame)
#' @param impfuncname The name of impurity function name.
#' @param majestic = 1, 2, 3, 4, or 5 to create a in-node resampling tree \cr
#' = -1 to create simple in-node resampling tree.
#' When majestic > 0, set seed prior to the run so that the result can be
#' reproduced.
#' @param method Treatment of the categorical regressor variable. \cr
#' = 0 to dummy expand the categorical regressor, \cr
#' = 1 to use the first canonical variable, \cr
#' = 2 to use up to the 2nd canonical variables and their sum and difference\cr
#' = 3 to use one-versus-rest dichotomization. \cr
#' = 4 to use dichotomization of the criterion variable on the basis of
#' all possible bipartitions of the levels of the criterion variable.
#' @param app = 1 or 9 to use all possible partitions
#' of categorical regressor levels. \cr
#' app=9 always uses all possible partitions regardless of ncaty.
#' @param maxapp Maximum # of categories to use app.
#' This will be used as the max # of categories of criterion
#' when method=4 is used.
#' @param maxncut Maximum # of cut point (threshold) candidates.
#' @param eps The minimum improvement of impurity measure.
#' @param minnobs The minimum # of observations in a node.
#' @param maxdepth The maximum depth of a tree.
#' @param maxdivide The maximum # of divisions.
#' @param minp minimum value of multinomial probability
#' @param nrepM # of divisions to try for majestic tree
#' @param psampleM percentage of the sample for majestic tree
#' @param rf = 1 to sample regressor variables (Random Forest)
#' @param nxvarM # of variables to sample for majestic tree (Random Forest)
#' @param .R.sM random number seed for majestic tree obtained as: \cr
#'  \code{.R.sM=get(".Random.seed", envir=globalenv() )} \cr
#'  in the previous run of this module.
#' @param minLeftM minimum proportion of x values to be classified as Left
#' @param maxRightM maximum proportion of x values to be classified as Right
#' @param minnobsM minimum # of obs in a node to use majestic
#' @param title The title string
#' @param print = 1 to print the result, = 2 to print the history
#' = 3 to print the history in detail.
#' @param plot = 1 to plot the tree.
#' @param debug = 1 to print intermediate result
#'
#' @details
#'
#' Input Variables: \cr
#' The criterion variable \code{y} must be a numeric vector or a factor.
#' The regressor variable \code{X} must be a data frame
#' with numeric or factor variables.
#'
#' Division Details \cr
#'
#' A Greedy Algorithm with Depth-First Search (DFS) is used to find a tree.
#'
#' For continuous regressor, all the unique values of it will be used
#' as the candidate for the threshold.
#'
#' Treatment of  categorical regressor variables: \cr
#' When \code{app=0}: \cr
#' The categorical regressor with the # of levels greater than 2
#' will be quantified according to \code{method}.
#'
#' Regardless of the method, if the criterion is continuous or
#' the # of levels of the criterion is equal to two,
#' the conditional mean of the criterion given the level of regressor
#' will be used as the quantified regressor. \cr
#' (The # of quantified regressor = 1.)
#'
#' If \code{method=0}  and if the # of levels of the criterion
#' is greater than 2, a categorical regressor will be dummy expanded
#' and the resulting dummy variables will be used as the quantified
#' regressor variables.
#' (The # of quantified regressor = # of levels of the regressor)
#'
#' If \code{method=1}
#' When the # of levels of the criterion is greater than 2,
#' the first canonical variable of the canonical discriminant analysis
#' of the criterion on the dummy expanded regressor will be used. \cr
#' (The # of quantified regressor = 1.)
#'
#' If \code{method=2}, in addition to the above, the 2nd canonical
#' variable, if exists, and the sum and difference of those two canonical
#' variables will be added as the quantified regressors. \cr
#' (The # of quantified regressors = 4 or 1.)
#'
#' If \code{method=3}, the one-versus-rest dichotomization of the criterion
#' variable is used and the conditional means will be used as the quantified
#' regressors.
#' If any of the resulting quantified variables has zero variance
#' or duplicated,
#' the redundant quantified variables will be removed. \cr
#' (The # of quantified regressors = 4 or 1 + # of levels of the criterion
#'  or less.)
#'
#' If \code{method=4},
#' all possible bipartitions of the levels of the criterion
#' is  used for the dichotomization
#' and the conditional means will be used as the quantified regressors. \cr
#' If any of the resulting quantified variables has zero variance
#' or duplicated,
#' the redundant quantified variables will be removed. \cr
#' (The # of quantified regressors = 2^{nc-1}-1 or less where nc is the
#' # of levels of the criterion variable.) \cr
#' This seems to be the best option when the # of levels of the criterion
#' is not so large ( <= 10 or so. ) and the # of levels of the regressor
#' is very large. \cr
#'
#'
#' When \code{app=1}: \cr
#'  If the levels of the criterion is greater than 2 and the levels of
#'  the regressor is greater than 2, all possible partitions of size 2
#'  will be tried.
#'
#' When \code{app=9}: \cr
#'  Regardless the # of levels of the categorical variables
#'  app will be used for the categorical regressors.
#'
#'
#' Stopping rule:\cr
#'  1) Relative improvement of the impurity measure due to a division
#'  is less than or equal to \code{eps} \cr
#'  2) # of observations in a node is less than or equal to \code{minnobs}.\cr
#'  3) # of observations in either subset becomes 0 after division. \cr
#'  4) Maximum # of division is greater than \code{maxdivide}. \cr
#'  5) Maximum depth of the tree reached \code{maxdepth}.
#'
#' Requires \code{igraph} package to draw a tree.
#'
#'
#' @examples
#' # univariate regression tree;
#' # numerical y and numerical X
#' X <- data.frame(x=univar$x)
#' plot( univar$x, univar$y1, type="p" )
#' res01 <- DecTree( univar$y1,X, minnobs=1, eps=0.001
#' , print=3, plot=0, app=0, title="y1" )
#' # library(igraph)
#' # plot_dt(res01$node)
#'
#' # categorical y and continuous X
#' res02 <- DecTree( as.factor(univar$y1),X, minnobs=1, eps=0.001, maxncut=100
#'                  , print=3, plot=0,, app=0, title="y1" )
#'
#' # categorical y and categorical X: (method 1 and 2 are not available)
#' Xc <- data.frame(x=as.factor(univar$x))
#' res03 <- DecTree( as.factor(univar$y1),Xc, minnobs=1, eps=0.001, method=3
#'                  , print=3, plot=0,, app=0, title="y1" )
#' res04 <- DecTree( as.factor(univar$y1),Xc, minnobs=1, eps=0.001, method=4
#'                  , print=3, plot=0,, app=0, title="y1" )
#'
#'
#'
#'
#' # bivariate classification tree
#' # categorical y, categorical X
#' SQ2=matrix(square$class4, 4,, byrow=1)
#' heatmap( SQ2, Rowv=NA, Colv=NA, reorderfun=NA, symm=1
#'        , labRow=1:4, labCol=1:4 )
#' Xc <- data.frame(x1=as.factor(square$x1),x2=as.factor(square$x2))
#' res0 <- DecTree( square$Class4, Xc, maxdepth=4, method=0, print=3
#'             , plot=0, title="square: categorical y, categorical X" )
#'
#' res1 <- DecTree( square$Class4, Xc, maxdepth=4, method=1, print=3
#'             , plot=0, title="square: categorical y, categorical X" )
#'
#' # categorical y, categorical X
#' res2 <- DecTree( square$Class4, Xc, maxdepth=4, method=2, print=3
#'           , plot=0, title="square: categorical y, categorical X" )
#'
#' # categorical y, categorical X
#' res3 <- DecTree( square$Class4, Xc, maxdepth=4, method=3, print=3
#'          , plot=0, title="square: categorical y, categorical X" )
#'
#' # categorical y, categorical X
#' res4 <- DecTree( square$Class4, Xc, maxdepth=4, method=4, print=3
#'          , plot=0, title="square: categorical y, categorical X" )
#'
#'
#' # categorical y, categorical X: all possible partitions of x levels
#' res11 <- DecTree( square$Class4, Xc, maxdepth=4, eps=0, app=1, print=3
#'                  , plot=0, title="square: categorical y, categorical X" )
#'
#' # categorical y, categorical X: all possible partitions of x levels
#' res12 <- DecTree( square$Class4, Xc,maxdepth=4, eps=0, app=9, print=3
#'                 , plot=0, title="square: categorical y, categorical X" )
#'
#'
#'
#' # categorical y, numerical X
#' X <- data.frame(x1=square$x1,x2=square$x2)
#' res21 <- DecTree( square$Class4, X, minnobs = 2, eps=0, print=3, plot=0
#'                    , title="square: categorical y, numerical X" )
#' plot2d_dt( square$Class4, X, res21$node, ngrid=50, title="Border Lines" )
#'
#'
#' # numerical y, numerical X
#' res22 <- DecTree( square$class4, X, minnobs=2, eps=0, print=3, plot=0
#'                  , title="square: numerical y, numerical X" )
#'
#'
#'
#' # slightly larger data
#' # categorical y, categorical X
#' Xc <- data.frame(x1=as.factor(square2$x1),x2=as.factor(square2$x2))
#' res31 <- DecTree( as.factor(square2$y), Xc, minnobs=50, eps=0.0001
#'                  , print=3, maxdepth=50, plot=0, app=9
#'                  , title="square: categorical y, categorical X" )
#'
#' # categorical y, categorical X
#' Xc <- data.frame(x1=as.factor(square2$x1),x2=as.factor(square2$x2))
#' res32 <- DecTree( as.factor(square2$y), Xc, minnobs=50, eps=0.0001
#'                  , print=3, maxdepth=50, plot=0, app=0, method=4
#'                  , title="square: categorical y, categorical X" )
#'
#'
#'
#'
#' # iris data
#' Species <- iris$Species
#' levels(Species) <- c("set","ver","vir")
#' ncaty <- 3; ncatX <- rep(0,4)
#' resI1 <- DecTree( Species, iris[,1:4], eps=1e-5
#'                    , title="rect: categorical y, continuous X", plot=0  )
#'
#'
#' # iris: categorical y, mixed X
#' Species <- iris$Species
#' levels(Species) <- c("set","ver","vir")
#' Xc <- iris[,-5]
#' Xc[,3] <- cut(Xc[,3],9, labels=1:9)
#' Xc[,4] <- cut(Xc[,4],9, labels=1:9)
#' ncaty <- 3; ncatX <- c(0,0,8,8)
#' resI2 <- DecTree( Species, Xc, eps=1e-5, app=1
#'                   , title="iris: categorical y, mixed X", plot=0  )
#' resI3 <- DecTree( Species, Xc, eps=1e-5, app=0, method=1
#'                   , title="iris: categorical y, mixed X", plot=0  )
#' resI4 <- DecTree( Species, Xc, eps=1e-5, app=0, method=3
#'                   , title="iris: categorical y, mixed X", plot=0  )
#' resI5 <- DecTree( Species, Xc, eps=1e-5, app=0, method=4
#'                   , title="iris: categorical y, mixed X", plot=0  )
#'
#'
#'
#'
#' @return
#' A class DecTree object containing
#' a list of the following:
#' \preformatted{
#' n: # of observations
#' ncaty: # of levels of y (0 means contimuous.)
#' levelsy: levels of y if y is categorical, or NULL
#' yname: the name of the criterion variable
#' yname2: the name of the actual y argument
#' nx: # of predictor variables
#' ncatx: # of levels of X (0 means contimuous.)
#' levelsX: levels of the regressor categorical variable, or NULL
#' xname: the names of predictor variables
#' xname2; the name of the actual X argument
#' impfuncname: the name of impurity function
#' nleaf: # of leaves
#' MaxDepth: The maximum depth of the tree
#' .R.s: .Random.seed vector
#'     Use this to reproduce the majestic tree.
#' RNGkindM, .R.sM: random number seed info
#'
#' node: A list of node information where node[[r]]  contains:
#'  node[[r]]$parent   The parent of the r-th node.
#'  node[[r]]$set      The set of observations in the r-th node.
#'  node[[r]]$nobs     # of observations in the r-th node.
#'  node[[r]]$j        The variable to be used to divide the r-th  node.
#'  node[[r]]$k     *  The quantified variable number to be used.
#'  node[[r]]$th    *  The value of the threshold or NA.
#'  node[[r]]$nqx   *  # of unique quantified variables.
#'  node[[r]]$valxj *  The set of factor levels of categorical variable j
#'                       to be classified as left from this node.
#'  node[[r]]$valxjR *  The set of factor levels of categorical variable j
#'                       to be classified as right from this node.
#'  node[[r]]$score    The nobs weighted impurity measure.
#'  node[[r]]$ImpMeasure score / nobs.
#'  node[[r]]$scoreimp The improvement of impurity measure due to dividing
#'                       this node.
#'  node[[r]]$impfuncname Name of the impurity measure.
#'  node[[r]]$depth    The distance from the top node
#'  node[[r]]$Left  *  The left node name
#'  node[[r]]$Right *  The right node name
#'
#'  node[[r]]$switched = 1 when left and right were swiched.
#'
#'  node[[r]]$yhat  ** The predicted value
#'  node[[r]]$prob  ** The post probability vector
#'
#'  * Not included in the final nodes.
#'  ** Not included in the middle nodes.
#'
#'
#' fimport: Feature Importance
#'
#' Tab: Cross Table
#' margin: Marginal Frequency
#' cor: Correlation Coefficient between continuous y and yhat.
#' Gini: gini coefficient
#' mllh: minus log multinomial likelihood
#' nparam: # of parameters estimated
#' aic and bic:
#' tprecision: total precision
#' macroF1 and weightedF1: macro F1 statistics (unweighted and weighted)
#' SCT: a matrix of accuracy, precision, recall, specificity, and F1-score
#'     for each levels of the criterion.
#'
#' args: list of arguments to DecTree function.
#' title: title string generated in DecTree function.
#'
#' y:  input
#' yhat: predicted value of y
#' prob: probability vector
#' Node: node number
#'
#' RNGkind: The name of randon mumber generator
#' .R.sM: The random number seed used for majestic.
#'
#' }
#'
#'
#'
#'
#' @references
#' Joe Suzuki (2020) Statistical Machine Learning with 100 Math and R
#' Problems. Kyoritsu Shuppan.
#'
#' @export
#'

DecTree <- function( y, X, impfuncname="gini", majestic=0, method=4, app=0
                , maxapp=12, maxncut=50, switch=0
                , eps=0, minnobs=1,maxdepth=10, maxdivide=500, minp=1e-4
                , nrepM=50, psampleM=100, rf=0
                , nxvarM=if(!is.null(y) && !is.factor(y))
                  max(floor(ncol(X)/3), 1) else floor(sqrt(ncol(X)))
                , .R.sM=NULL
                , minLeftM=0.5, maxRightM=0.5, minnobsM=10
                , title="", print=1, plot=0, debug=0 ){
  # Modified version of "df" function in Joe Suzuki (2020)
  # Shin-ichi Mayekawa
  # 20230901
  # return threshold from branch: 20230902cot,cot,03,04,05,06
  # makeStack: 20230908
  # dt --> DecTree, x --> X: 20230909
  # scalex: 20230909
  # X,y --> y,X: 20230909
  # maxcat, call print_hist, cross_table, plot_dt: 20230909
  # store categorical variables as factors: 20230910
  # detect ncat and ncatx: 20230910
  # print_leaf: 20230910
  # method: 20230911
  # makeStack moved into DecTree: 20230912
  # x < th  -->  x <= th: 20230914
  # k --> r,  stackr -->  stacktop: 20230914
  # scoreimpr: 20230915
  # cross table w/o classify: 20230916
  # unweighted impurity function: 20230916
  # feature importance: 20230917
  # depth: 20230917
  # leaf node info in while loop: 20230918
  # center --> yhat: 20230918
  # Left and Right: 20230918
  # instead of (y,X), pass (ys, Xs) to Divide: 20230920
  # centropy: 20230925
  # app (all possible partitions): 20230926dnc,27
  # maxapp: 20230927
  # plot type=2: 20230927
  # alpha --> eps, n.min --> minobs,  vertex --> node: 20230928
  # sq.loss --> variance: 20230930
  # plot=0: 20231001
  # ncatx length(levels) -> unique: 20231002
  # output prob: 20231002
  # classify -->  predict_dt: 20231002
  # print_hist, print_leaf --> print_..._dt: 20231002
  # errCol: 20231003
  # expand categorical  y: 20231004,05
  # roxygen2 fix: 20231014
  # bugfix for importance: 20231014,15
  # names added to node: 20131015
  # inpfuncname -> impfuncname: 20231015
  # output nobs of leaf node, ImpMeasure, impfuncname: 20231015
  # class DecTree: 20231016
  # output xname etc: 20231016
  # set method=4 as default: 20231014mm
  # print, plot > 0: 20231102
  # new method=3: 20231108
  # output nqx: 20231109
  # llh: 20231109
  # use of summary_dt: 20231109
  # new title: 20231110cot,20231121dnc
  # output weightedF1: 20231112cot
  # return args: 20231121dnc
  # ncat <= maxapp when method=4: 20231121dnc
  # levels: 20231122
  # output y and yhat: 20231123
  # output Gini: 20231125
  # majestic: 20231223cot,27,28,29
  # definition of ncatx: 20231223
  # print_leaf_dt, print_hist_dt and Mode made independent: 20231228
  # minobs -> minnobs: 20231229
  # plotM: 20240105,09dnc
  # new plot: 20240110
  # switch: 202430115
  # valxjR: 20240211
  # output levelsX: 20240216dnc
  # MaxDepth: 20240218
  # scoreimpr < eps  -->  scoreimpr <= eps etc: 20240224
  # new convergence crit: 20240225
  # switch=2: 20240226,27dnc
  # remove empty nodes: 20240227dnc
  # majestic to DivedeM: 20240228
  # RNGkindM: 20240309
  # treatment of .Random.seed: 20240309
  # set seedM here and return .R.sM: 20240313
  # return .R.sM when seedM=0: 20240314,15
  # simpler version with doRNG: 20240318,19,20
  # save and restore the original seed: 20240326
  # new stopping rule: 20240328dnc
  # seedM, plotM removed: 20240330
  # majestic=3 comment: 20240407cot
  # print majestic options: 20240409
  # majestic=-1: 20240416,17
  # majestic, nrepM added to class DecTree: 20240417
  # default minnobsM=10: 20240423
  # RF: 20240425,27
  # RF=0 as default: 20240502
  # return yhat as factor: 20240502
  # bugfix for majestic=-1 .R.sM: 20240507
  #

  #
  # Original version named as dt is found in: Joe Suzuki (2020)
  #


  # variable type by # of categories
  # ncat=0 or 1  continuous
  # ncat=2 binary
  # ncat > 2 polytomous
  #



  # check data
  if( !is.data.frame(X) ){
    cat("\nerror1:(DecTree) X must be a data frame.\n")
    stop()
  }
  if( any( unlist(lapply(X,is.character)) ) ){
   cat(
    "\nerror1:(DecTree) Character variables must be converted to factors.\n")
    stop()
  }
  if( is.character(y) ){
  cat(
    "\nerror1:(DecTree) Character variables must be converted to factors.\n")
    stop()
  }
  # missing data
  if( any(is.na(y)) || any(is.na(X)) ){
   cat("\nerror1:(DecTree)  No missing data is allowed.\n")
    stop()
  }


  # args
  args=named_list( impfuncname, method, app, maxapp, maxncut, eps, minnobs
                  , maxdepth, maxdivide, minp, title, print, plot, debug
                  , majestic, nrepM, rf, psampleM, nxvarM, .R.sM, minLeftM
                  , maxRightM, minnobsM, switch )

  # # of obs and regressors
  n <- length(y)
  nx=ncol(X)
  xname=colnames(X)
  xnamen=paste(1:nx,xname,sep=":")


  xname2=as.character(substitute(X))
  yname2=as.character(substitute(y))
  yname=names(y)
  levelsy=levels(y)
  if( length(yname2) > 1 ){
    yname2=paste(yname2[2],yname2[1],yname2[3], sep="")
  }

  # scale level of y and impurity measure
  if( is.factor(y) ){
    ncaty=length(levels(y))
    ncaty=length(unique(y))                    # 20231004
    if (impfuncname %in% c("mis.match","missmatch", "mismatch") ){
      g <- mis.match
    } else if (impfuncname == "gini"){
      g <- gini
    } else{
      g <- entropy
    }
    ImpFunc=impfuncname
  }
  else{
    ncaty=0
    g=variance
    ImpFunc="variance"
  }

  # Note that, ncaty<=1 means y is continuous,
  # ncaty>=2 means y is categorical.
  # y is categorical when is.factor(y)=TRUE.
  # The same stands for each column of X.

  # scale levels of X
  ncatx=rep(0,nx)
  for( j in 1:nx ){
    # if( is.factor(X[,j]) ) ncatx[j]=length(unique(X[,j]))
    if( is.factor(X[,j]) ) ncatx[j]=length(levels(X[,j]))    # 20231223
  }
  names(ncatx)=xname
  levelsX=lapply( X, levels )

  # expand y if ncaty is not too large and there are categorical x
  # The resulting  Y is a binary matrix of size n x 2^{ncaty-1}-1
  # whose columns correspond to the method of dichotomization of y.
  # The first ncaty columns correspond to one and the rest method.
  # The remaining columns correspond to the partitioning of ncaty levels
  # such that (1,2) and the rest, (1,3) and the rest, so on.
  #
  # For 1 <= j <= ncat,
  #  Y[i,j] == 1 means y[i] == unique(y)[j]
  # For j == ncat+1
  #  Y[i,j] == 1 means y[i] == unique(y)[1]  or y[i] == unique(y)[2]
  # For j == ncat+2
  #  Y[i,j] == 1 means y[i] == unique(y)[1]  or y[i] == unique(y)[3]
  #
  if( method >= 3  &&  ncaty >= 3  && ncaty <= maxapp  &&  any(ncatx >= 3) ){

   id=gen01pat( ncaty, sort=-2, partition=1 ) == 1  # logical matrix
   if( method == 3 ) id=id[1:ncaty,,drop=0]

   npart=nrow(id)
   uy=unique(y)
   Y=matrix(0,n,npart)
   for(k in 1:npart ){
     Y[,k]=y %in% uy[id[k,]]
   }

  } # end of Y for method=4
  else{
   Y=matrix(0,n,1)  # to take care of subsetting
  }

  if( rf == 0 ){
    nxvarM=nx
  }
  else{
    nxvarM=round(nxvarM)
  }

  title0=paste("[",ImpFunc, "]",  " ncaty=",ncaty
               , ", app=",app, "(",maxapp,")"
               , ", method=",method, "(",maxncut,")"
               , ", maxdepth=",maxdepth, sep="")
  nchar0=nchar(title0)
  nchar1=nchar(title)
  if( nchar0+nchar1 > 60 ) title=paste(title,"\n",title0)
  else if( nchar1 > 0 ) title=paste(title,":",title0)
  else title=title0

  if( print > 0 ){
    cat("\nDecision Tree Analysis\n")
    cat("\n Node Impurity Measure =", ImpFunc, "\n" )
    cat(" method =", method, ",  app =", app, ",  maxapp =", maxapp
        , ", maxncut =", maxncut, "\n")
    Print( maxdivide, maxdepth, eps, minnobs )
    cat( " majestic = ", majestic, "\n" )
    if( majestic != 0 ){
      cat( " nrepM = ", nrepM, ", rf = ", rf, ", nxvarM = ", nxvarM
           , ", psampleM = ", psampleM
           , ", minnobsM =", minnobsM, "\n"
      , " , minLeftM = ", minLeftM, ", maxRightM = ", maxRightM, "\n" )
      if( !is.null(.R.sM) ) cat(" .R.sM is given.\n")
    }
    Print(n,nx)
    Print(ncaty, xnamen, ncatx)

    if( print >= 3 ){
      cat("\n",title,sep="")
      cat("\nDivision History Details\n")
    }
  }

  #
  # Greedy Algorithm
  # Depth-First Search (DFS) using stack.
  #

  #############################################
  # prepare stack
  STACK=makeStack()
  PUSH <- STACK$push
  POP <- STACK$pop
  LEN <- STACK$length
  # INSP <- STACK$inspect
  rm(STACK)
  #############################################


  #
  # If .R.sM is defined, use it as the seed for majestic.
  # Otherwise, save the current seed in .R.sM and return it upon exit.
  #

  # save the current seed
  .R.s=get(".Random.seed", envir=globalenv() )

  if( majestic  &&   !is.null(.R.sM) ){           # 20240507
    # use .R.sM as the random seed
    assign( ".Random.seed", .R.sM, envir=globalenv() )
  }
  RNGkindM=RNGkind()



  # initialize stack
  PUSH( list( parent=0, set=1:n, score=length(y)*g(y), depth=0 ) )

  node=list()
  r=0
  rr=0
  nleaf=0
  MaxDepth=0
  jp=0

  while( LEN() > 0 ){

    # current node number
    # node r consists of the observations stored in stacktop$set.
    r=r+1

    # get next set to be processed from the stack.
    stacktop=POP()

    pa=stacktop$parent
    nobs=length(stacktop$set)
    score=stacktop$score
    depth=stacktop$depth
    set=stacktop$set

    if( print >= 3)
     cat(" ", r, "  pa=", pa, ", nobs=", nobs
         , ", score=", round(score,3),", ", sep="")


    # stacktop$parent   The parent number of current set
    # stacktop$set      The set of observations
    # stacktop$score    The nobs weighted impurity score
    #                   when this set was created.
    #                   Score is needed to check the convergence.
    #                   When y is continuous, score is the Total SS.
    # stacktop$depth    Distance from the top node (depth)


    # divide the current node into left and right.

    if( majestic && nobs >= minnobsM ){
     res <- DivideM( y[stacktop$set], X[stacktop$set,,drop=0]
                    , Y[stacktop$set,,drop=0], impfunc=g
                    , S=stacktop$set, method=method, ncatx=ncatx
                    , app=app, maxapp=maxapp, maxncut=maxncut
                    , levelsy=levelsy, levelsX=levelsX, majestic=majestic
                    , psampleM=psampleM, nxvarM=nxvarM, nrepM=nrepM, rf=rf
                    , minLeftM=minLeftM, maxRightM=maxRightM, r=r
                    , debug=max(0,debug-3) )
    }
    else{
     res <- Divide( y[stacktop$set], X[stacktop$set,,drop=0]
                   , Y[stacktop$set,,drop=0], impfunc=g
                   , S=stacktop$set, method=method, ncatx=ncatx
                   , app=app, maxapp=maxapp, maxncut=maxncut
                   , levelsy=levelsy, levelsX=levelsX )
    }

    # res$j     the variable number used to divide the set
    # res$threshold the threshold of variable j when it is continuous
    # res$valxj the set of values of nominal variable j to be classified
    # as left
    # res$valxjR the set of values of nominal variable j to be classified
    # as right
    # res$k     quantified variable number used.
    # res$nqx   # of unique quantified variable
    # res$left  the set of observations classified as left
    # res$right the set of observations classified as right
    # res$left.score  The nobs weighted impurity measure of left set
    # res$right.score The nobs weighted impurity measure of right set
    # res$score       The sum of above two.
    # res$Mrules
    #
    #

    # Below, node[[r]] will contain:
    #  node[[r]]$parent   The parent of the r-th node
    #  node[[r]]$set      The set of observations in the r-th node
    #  node[[r]]$nobs     # of obs in the r-th node
    #  node[[r]]$j        The variable to be used to divide this node
    #  node[[r]]$k        The quantified variable number used.
    #  node[[r]]$th       The value of the threshold.
    #  node[[r]]$valxj    The set of factor levels of categorical
    #                     regressor j to be classified as left.
    #  node[[r]]$valxjR   The set of factor levels of categorical
    #                     regressor j to be classified as right.
    #  node[[r]]$nqx      # of quantified variables of regressor j
    #  node[[r]]$score    The nobs weighted score
    #  node[[r]]$scoreimp The improvement of nobs weighted score
    #                     due to dividing this node.
    #  node[[r]]$depth    The distance from the top node
    #  node[[r]]$Left     The left node name
    #  node[[r]]$Right    The right node name (to be added later)
    #  node[[r]]$impfuncname Name of the impufity function used
    #  node[[r]]$ImpMeasure The nobs weighted score / nobs
    #
    #  node[[r]]$switched = 1 if the right and the left are switched
    #
    #  node[[r]]$yhat
    #  node[[r]]$prob
    #
    #  node[[r]]$Mrules
    #

    # change integer(0) to NULL (may be the result of native which function.)
    if( length(res$valxj) == 0 ) res$valxj=NULL
    if( length(res$valxjR) == 0 ) res$valxjR=NULL


    ##########################################################################
    ##########################################################################
    ##########################################################################
    # switch nodes according to obs number
    #
    # When switched==1, those x satisfying x < th will be classfied as RIGHT.
    #
    switched=0

    if( switch == 1 && length(res$left)*length(res$right) > 0 ){
      statL=summary(res$left, quantile.type=1)
      statR=summary(res$right, quantile.type=1)
      statd=statL < statR
      #   Print(r,pa, res$left.score, res$right.score,statL, statR )
      # hist(res$left, main=paste(r," left"))
      # hist(res$right, main=paste(r," right"))

#     if( sum(statd) < 3 ){
      if( !statd["Mean"] ){
          # switch so that statL <= statR
 #       Print(r,res$j, res$k, res$threshold)

        switched=1
        left=res$right
        right=res$left
        left.score=res$right.score
        right.score=res$left.score
        if( ncatx[res$j] > 0 ){
          valxj=res$valxjR
          valxjR=res$valxj
        }
        else{
          valxj=NULL;  valxjR=NULL
        }

        res$left=left
        res$right=right
        res$left.score=left.score
        res$right.score=right.score
        res$valxj=valxj
        res$valxjR=valxjR
      }
    }
    else if( switch == 2 && length(res$left)*length(res$right) > 0 ){
      if( !is.null(res$valxj) && !is.null(res$valxjR) ){
        statL=summary(res$valxj, quantile.type=1)
        statR=summary(res$valxjR, quantile.type=1)
        statd=statL < statR
        #   Print(r,pa, res$left.score, res$right.score,statL, statR )
        # hist(res$left, main=paste(r," left"))
        # hist(res$right, main=paste(r," right"))

        #     if( sum(statd) < 3 ){
        if( !statd["Mean"] ){
          # switch so that statL <= statR
          #       Print(r,res$j, res$k, res$threshold)

          switched=1
          left=res$right
          right=res$left
          left.score=res$right.score
          right.score=res$left.score
          if( ncatx[res$j] > 0 ){
            valxj=res$valxjR
            valxjR=res$valxj
          }
          else{
            valxj=NULL;  valxjR=NULL
          }

          res$left=left
          res$right=right
          res$left.score=left.score
          res$right.score=right.score
          res$valxj=valxj
          res$valxjR=valxjR
        }
      }
    } # end of switch


    ##########################################################################
    ##########################################################################
    ##########################################################################


    # improvement of impurity measure
    # score is Total SS of node[r].
    # res$score is the sum of within SSs after dividing node[r].
    # Therefore, scoreimp=score-res$score is the between SS.

    if( debug > 0 ){
      cat("\n After Divide")
      Print(res$left.score, res$right.score, res$score)
    }

    if( score > 0 ){
      scoreimp=(score-res$score)
      scoreimpr=scoreimp/score
    }
    else{
      scoreimp=0
      scoreimpr=0
    }
    ImpMeasure=score/nobs

    nobsL=length(res$left)
    nobsR=length(res$right)

    if( print >= 3)
      cat( "scoreimpr=", round(scoreimpr,3), ", j=", res$j, ", k=", res$k
           , ", th=", round(res$threshold,3)
           , ", nobsL R=", nobsL, " ", nobsR, sep="","\n" )

    # check if continue or not

    if( debug > 0 ){
      cat("\n Before convergence check.")
      Print(score, scoreimp, nobs, nobsL, nobsR, r,depth)
    }

    if( debug ){
     Print( r, depth, score, scoreimpr, nobs, res$j, jp)
    }

    if( score == 0   || (nobs <= minnobs && scoreimpr <= eps)
        || (score != 0  &&  nobsL*nobsR == 0)
        || nobs <= minnobs
        || r >= maxdivide || depth >= maxdepth ){

      # leaf node: stop dividing
      # Store the leaf node info in node[[r]] and no need to PUSH.
      #
      if( ncaty == 0 ){
        yhat=mean(y[stacktop$set])
        prob=NA
      }
      else{
        yhat=Mode(y[stacktop$set])
        prob=table(y[stacktop$set])
        prob=prob/sum(prob)
      }

      nleaf=nleaf+1

      node[[r]] <- list( parent=stacktop$parent, set=stacktop$set, j=0
                          , k=0, nqx=0
                          , depth=depth, yhat=yhat, prob=prob, nobs=nobs
                          , score=score, ImpMeasure=ImpMeasure, Left=0
                          , impfuncname=impfuncname, switched=switched )
      if( majestic < 0 ) node[[r]]=append( node[[r]], list(Mrules=res$Mrules) )

    }
    else{

     # middle node: keep dividing
     #
     # store the middle node info in node[[r]]
     node[[r]] <- list( parent=stacktop$parent, set=stacktop$set
                        , nobs=nobs, Left=r+1
                        , j=res$j, th=res$threshold, k=res$k, nqx=res$nqx
                        , valxj=res$valxj, valxjR=res$valxjR
                        , scoreimp=scoreimp, score=score, depth=depth
                        , impfuncname=impfuncname, ImpMeasure=ImpMeasure
                        , switched=switched )
     if( majestic < 0 ) node[[r]]=append( node[[r]], list(Mrules=res$Mrules) )

     # push right first so that left will be popped first.
     PUSH( list(parent=r, set=res$right, score=res$right.score
                , depth=depth+1) )
     PUSH( list(parent=r, set=res$left, score=res$left.score
                , depth=depth+1) )
    }

 # Print( r, LEN(), stacktop, node[[r]] )

    # depth
    MaxDepth=max(MaxDepth,depth)

    jp=res$j

  } # end of while


  # final treatment of node
  # add right node name
  m=length(node)
  for( r in 1:m ){
    node[[r]]$Right=0
  }
  for( r in m:2 ){
    pa=node[[r]]$parent
    if( node[[pa]]$Right == 0 ){
      node[[pa]]$Right=r
    }
  }
  names(node)=1:length(node)

  #
  #
  # Here, node is ready.
  #
  #

  if( majestic > 0  &&   !is.null(.R.sM) ){
   # restore original seed
   assign( ".Random.seed", .R.s, envir=globalenv() )
  }



  # remove empty nodes, if any.   20240227dnc
  # There is a slight chance that DivideM may create empty nodes.
  # Otherwise, not sure why there exist empty nodes.
  #
  Nodes=1:m
  Nobs=get_valLL( node, "nobs", method=2 )
  Parent=get_valLL( node, "parent", method=2 )
  Left=get_valLL( node, "Left" )
  Right=get_valLL( node, "Right" )

  loczero=which(Nobs == 0)
  if( length(loczero) > 0 ){

   # remove zero-obs nodes
#   Print(Nodes,Nobs,Parent,Left,Right)

   from=Nodes[-loczero]
   to=1:length(from)
   Parent=recode(Parent,from,to)
   Left=recode(Left,from,to)
   Right=recode(Right,from,to)

   for( i in 1:m ){
    node[[i]]$parent=Parent[[i]]
    node[[i]]$Left=Left[[i]]
    node[[i]]$Right=Right[[i]]
   }
   node=node[-loczero]
   m=length(node)
   names(node)=1:m

#   Nodes=1:m
#   Nobs=get_valLL( node, "nobs" )
#   Parent=get_valLL( node, "parent" )
#   Left=get_valLL( node, "Left" )
#   Right=get_valLL( node, "Right" )
#   Print(Nodes,Nobs,Parent,Left,Right)

  }


  # summary stat
  temp=summary_dt( node, y, print=0, outpred=1, nleaf=nleaf )
# Print("after summary DecTree")
  Tab=temp$Tab
  margin=temp$margin
  cor=temp$cor
  Gini=temp$Gini
  mllh=temp$mllh
  nparam=temp$nparam
  aic=temp$aic
  bic=temp$bic
  tprecision=temp$tprecision
  macroF1=temp$macroF1
  weightedF1=temp$weightedF1
  SCT=temp$SCT
  yhat=temp$yhat
  if( ncaty > 0 ) yhat=factor(yhat,levels=levelsy)   # 20240502
  prob=temp$prob
  Node=temp$Node
  MaxDepth=MaxDepth


  # feature importance
  nobs=as.numeric( get_valLL(node,"nobs")[,"_v_1"] )
  var=as.numeric( get_valLL(node,"j")[,"_v_1"] )
  scoreimp=as.numeric( get_valLL(node,"scoreimp")[,"_v_1"] )
  locnz=which(var != 0)
  nobs=nobs[locnz]
  var=var[locnz]
  scoreimp=scoreimp[locnz]
  # importance=scoreimp/nobs # 20231005
  importance=scoreimp
  # Print(var,nobs, importance)

  if( majestic >= 0 ){
    temp=xtabs(importance~var)
    temp=temp/sum(temp)
    # Print(temp,names(temp))
    fimport=numeric(nx)
    fimport[as.numeric(names(temp))]=temp
    names(fimport)=xname
  }
  else{
    fimport=NA
  }
  if( print > 0 ){

   print_hist_dt( node, title )

   cat("\nTotal # of leaves =", nleaf, "\n")
   if( is.factor(y) ){
     cat("\n", title, sep="")
     cat("\nConfusion Matrix from Training Data:  row=true, col=predicted\n")
     Print(Tab, margin)
     cat("\nFit Statistics\n")
     Print(SCT)
     cat("* recall = True Positive Rate,  specificity = True Negative Rate\n")
     Print(mllh, nparam, aic, bic)
     Print(Gini, tprecision, macroF1, weightedF1, fmt=".4 7.4")
   }
   else{
    cat("\n", title, sep="")
    cat("\nCorrelation Coefficient between y and yhat\n")
     Print(cor)
     Tab=NA; margin=NA
   }

   if( majestic >= 0 ){
    cat("\nFeature Importance\n")
    Print(fimport, fmt="7.3")
   }

  }

#Print(node)

  if( plot > 0 ){
   if( ncaty > 0 ){
    plot_dt( node, title, xname=xname )
   }
   else{
    plot_dt( node, title, type=2, xname=xname )
   }
  }


  res=named_list( n, ncaty, levelsy=levelsy, yname, yname2
                  , nx, ncatx, xname, xname2, levelsX=levelsX
                  , impfuncname, nleaf, MaxDepth, majestic, nrepM
                  , node, fimport, Tab, margin, cor
                  , Gini, mllh, nparam, aic, bic
                  , tprecision, macroF1, weightedF1, SCT
                  , args, title, y, yhat, prob, Node
                  , RNGkindM, .R.s=.R.s )
  class(res)="DecTree"

 # if( print > 0 )  print_leaf_dt( res, title )

  return(res)


} # end of DecTree




