#' Impurity Measure Functions
#'
#' gini, entropy, mis.match are for categorical criterion and
#' variance is for numerical criterion. \cr
#' Japanese help file: \link[lazy.tree]{gini_JPH}
#'
#'
#' @param y The input vector (numeric)
#'
#' @details
#' These impurity measures are not weighted by the # of observations
#' in the set.
#'
#' \code{gini} Gini impurity measure \cr
#' \code{entropy} cross entropy impurity measure \cr
#' \code{mis.match} proportion of observations whose predicted value
#' is not the same as the mode, \cr
#' \code{variance} variance of the criterion in the subset.
#' This is equal to the resudial sum of squares divided by n. \cr
#'
#' \code{entropy} can be interpreted as minus the log likelihood of
#' the multinomial distribution.
#'
#' @export
#'

gini <- function( y ){
  # gini impurity measure
  # Shin-ichi Mayekawa
  # 20230314,0904,05
  #
  if (length(y) == 0) return(0)
  freq=as.numeric(table(y))
  p=freq/sum(freq)
  gini=sum(p*(1-p))
  return( gini )

} # end of gini


#' @rdname gini
#' @export
#'

centropy <- function( y ){
  # gini impurity measure
  # Shin-ichi Mayekawa
  # 20230925
  #
  if (length(y) == 0) return(0)
  freq=as.numeric(table(y))
  p=freq/sum(freq)
  ce=-sum(p*log(p + 1e-9))
  return( ce )

} # end of cross entropy



#' @rdname gini
#' @export
#'

mis.match <- function(y) {
  if (length(y) == 0) return(0)
  return( sum(y != Mode(y))/length(y) )
} # end of mis.match


#' @rdname gini
#' @export
#'

entropy <- function(y) {
  # This is equal to multinomial likelihood divided by n
  # when the sample proportions of each category is used
  # as the predicted values of one-hot dummy variables:
  #  sum( f log(phat) ) = sum( p log(phat) ) * n
  if( length(y) == 0 ) return(0)
  f=as.numeric(table(y))
  locnz=which(f > 0)
  p=f[locnz]/sum(f)
  return(-sum(p*log(p)))
} # end of entropy





#' @rdname gini
#' @export
#'
variance <- function(y) {
  # rss / n = variance
  # sq.loss renamed as variance: 20230930
  if (length(y) == 0) return(0)
  return( sum( (y-mean(y))^2 )/length(y) )
} # end of variance




