#' Plot the Border Lines Calculated by the Decision Tree
#'
#' @param y The categorical criterion varialbe as factor
#' @param X The bivariate numeric predictor matrix (data frame)
#' @param obj The class DecTree or list of nodes.
#' @param ngrid # of grid points to be used
#' @param title The tile string
#'
#'
#' @examples
#' # eample 1 with sample data square
#' SQ2=matrix( square$class4, 4,, byrow=1 )
#' heatmap( SQ2, Rowv=NA, Colv=NA, reorderfun=NA, symm=1
#'       , labRow=1:4, labCol=1:4 )
#' X=as.data.frame(cbind(square$c1,square$c2))
#' res1 <- DecTree( square$Class4, X, maxdepth=5, print=1 )
#' plot2d_dt( square$Class4, X, res1$node, ngrid=50, title="square" )
#'
#' # example 2 with sample data rect
#' X=as.data.frame(cbind(rect$x1,rect$x2))
#' res2 <- DecTree( as.factor(rect$R), X, maxdepth=5, print=1, method=4 )
#' plot2d_dt( as.factor(rect$R), X, res2$node, title="square" )
#'
#'
#' @references
#' http://lbm.ab.a.u-tokyo.ac.jp/~omori/kensyu/discriminant.htm
#'
#' @export
#'

plot2d_dt <- function( y, X, obj, ngrid=50, title="border lines" ){
  # plot the boorder lines
  # Shin-ichi Mayekawa
  # 20231002,03
  # check if X is numeric: 20231007cot
  #
  # Original version was in:
  # http://lbm.ab.a.u-tokyo.ac.jp/~omori/kensyu/discriminant.htm
  #
  # DecTree object: 20231016
  #

  # DecTree object
  if( class(obj) == "DecTree" ){
    node=obj$node
    rm(obj)
  }
  else if( is.list(obj) & !is.data.frame(obj) ){
    nam=names(obj[[1]])
    if( all(
     c("parent","set","depth","nobs","Left","Right","j") %in% names(obj[[1]])
    ) ){
      node=obj
      rm(obj)
    }
  }
  else{
    cat("\nerror1:(plot2d_dt) The input object is not a DecTree object.\n")
    stop()
  }


 if( !is.factor(y) ){
  cat("\nerror1:(plot2d_dt) The criterion must be categorical. \n")
  stop()
 }
 if( any(!unlist( lapply(X,is.numeric) )) ){
  cat("\nerror1:(plot2d_dt) The regressors must be numeric. \n")
  stop()
 }
 cname=colnames(X)
  ncaty=length(unique(y))

  min1=min(X[,1])-0.5; max1=max(X[,1]+0.5)
  min2=min(X[,2])-0.5; max2=max(X[,2]+0.5)
  x1 <- seq( min1, max1, length=ngrid )
  x2 <- seq( min2, max2, length=ngrid )
  grid <- expand.grid(x1,x2)
  prob <- predict_dt( grid, node )$prob

  plot( X, type="n", xlim=c(min1,max1), ylim=c(min2,max2),
       xlab=cname[1], ylab=cname[2], main=title )
  text( X, as.character(y), pos=2 )

  colors=c("black","red","green","blue","magenta")
  for( i in 1:ncaty ){
   zp <- prob[,i] - apply(prob[,-i],1,max)
   contour(x1, x2, matrix(zp,ngrid), add=T, levels=0, col=colors[i] )
  }
  for( i in ncaty-(1:ncaty)+1 ){
     zp <- prob[,i] - apply(prob[,-i],1,max)
     contour(x1, x2, matrix(zp,ngrid), add=T, levels=0, col=colors[i] )
  }

} # end of plot2d_dt



