#' Calculate the Predicted Values from the Regressor Matrix
#'
#' This function classifies the observations into one of the leaf nodes. \cr
#' Japanese help file: \link[lazy.tree]{predict_dt_JPH}
#'
#' @param X The regressor matrix as a data frame
#' @param obj The class DecTree object or list of nodes.
#'
#'
#' @details
#' 
#' The rules concerning how to classify an observation into Left and Right node
#' recorded in each \code{obj$node} will be applied for each
#' observations (rows) in \code{X}.
#' 
#' When \code{X} contains a new set of observations with categorical
#' regressor variables, those must be converted to factors using the same
#' levels as the training data.
#'
#' @examples
#' Xc=data.frame(square$c1,square$c2)
#' res1 <- DecTree( square$Class4, Xc, maxdepth=3, print=1 )
#' pred <- predict_dt( Xc, res1 )
#' Print(pred$yhat, pred$prob)
#'
#' @return
#' A list of: \cr
#'  yhat: vector of the predicted values \cr
#'  prob: matrix of the posterior probabilities \cr
#'  Node: vector of the leaf node numbers
#'
#' @export
#'

predict_dt <- function( X, obj ){
  # get predicted values from u vector and node:  with categorical regressor
  # Shin-ichi Mayekawa
  # 20230906
  # convert factor to numeric: 20230910
  # x < th -->  x <= th: 20230914
  # Left and Right: 20230918
  # vertex --> node: 20230928
  # output prob: 20231002
  # classify -->  predict_dt: 20231002
  # DecTree object: 20231016
  # roxygen2: 20231024mm
  # output Node: 20231125
  # majestic < 0: 20240417
  #


  aggregate_dt <- function( Xs, ncatx, nrepM, Mrules ){
    # aggregate majestic<0 result
    # returns sumyhat5=1 if the obs is classfied as Left
    #
    nobs=nrow(Xs)

    sumyhat=numeric(nobs)
    yhat=numeric(nobs)

    for( rep in 1:nrepM ){
      yhat=numeric(nobs)

      j=Mrules$Jall[rep]
      if( ncatx[j] <= 2 ){
        yhat[ as.numeric(Xs[j]) <= Mrules$THall[rep] ]=1
      }
      else{
        Vj=Mrules$Valxjall[[rep]]
        for( a in 1:length(Vj) ){
          yhat[ as.numeric(Xs[j]) == Vj[a] ]=1
        }
      }

      if( rep == 1 ) sumyhat=yhat
      else{
        if( Mrules$Reflect[rep] ) yhat=1-yhat
        sumyhat=sumyhat+yhat
      }

    } # end of rep loop

    # aggregated yhat:  Left or Right
    sumyhat=sumyhat/nrepM
    sumyhat5=as.factor(as.numeric(sumyhat >= 0.5))

    return( sumyhat5 )

  } # end of aggregate_dt



  # DecTree object
  if( class(obj) == "DecTree" ){
    node=obj$node
    majestic=obj$majestic
    nrepM=obj$nrepM
    ncatx=obj$ncatx
    minnobsM=obj$args$minnobsM
    rm(obj)
  }
  else if( is.list(obj) & !is.data.frame(obj) ){
    nam=names(obj[[1]])
    if( all(
     c("parent","set","depth","nobs","Left","Right","j") %in%
     names(obj[[1]]) ) ){
      node=obj
      majestic=0
      nrepM=1
      ncatx=NA
      minnobsM=NA
      rm(obj)
    }
  }
  else{
    cat("\nerror1:(predict_dt) The input object is not a DecTree object.\n")
    stop()
  }


  ###############################################################
  #
  # Note that X must be factorized using the same levels as
  # the training data set which created the node.
  # 20230919
  #
  ###############################################################


  nobs=1
  if( is.data.frame(X) || is.matrix(X) ) nobs=nrow(X)
  else X=matrix(X,1)

  ncaty=length( node[[length(node)]]$prob )
  if( ncaty == 1 ) ncaty=0

  class=numeric(nobs)
  Node=numeric(nobs)
  if( ncaty >= 2 ) prob=matrix(0,nobs,ncaty)
  else prob=NA

  # make X numeric, not factor,  because divide uses numeric X
  if( is.data.frame(X) ) X[]=lapply(X,as.numeric)
  else if( is.matrix(X) ) X[]=apply(X, 2,as.numeric)
  else X=as.numeric(X)


  if( majestic >= 0 ){

    # majestic >= 0

    # for each row of X
    for( i in 1:nobs ){

      u=X[i,]

      r <- 1
      while (node[[r]]$j != 0) {

        if( is.null(node[[r]]$valxj) ){
          # continuous predictor
          # compare the value of u[j] to the r-th threshold
          if (u[node[[r]]$j] <= node[[r]]$th) left=1
          else left=0
        }
        else{
          # categorical predictor
          # check if the value of u[j] is in the categories of xj
          if (u[node[[r]]$j]  %in% node[[r]]$valxj) left=1
          else left=0
        }

        if( left == 1){
          r <- node[[r]]$Left
        } else {
          r <- node[[r]]$Right
        }

      } # end of while

      Node[i]=r
      class[i]=node[[r]]$yhat
      if( ncaty >= 2 ) prob[i,]=node[[r]]$prob

    } # end of i loop for obs


  } # end of majestic >= 0
  else{

    # majestic < 0

    # for each row of X
    for( i in 1:nobs ){

      u=X[i,]

      r <- 1
      while (node[[r]]$j != 0) {

        if( node[[r]]$nobs >= minnobsM ){
          left=aggregate_dt( u, ncatx, nrepM, node[[r]]$Mrules )
        }
        else{
          if( is.null(node[[r]]$valxj) ){
            # continuous predictor
            # compare the value of u[j] to the r-th threshold
            if (u[node[[r]]$j] <= node[[r]]$th) left=1
            else left=0
          }
          else{
            # categorical predictor
            # check if the value of u[j] is in the categories of xj
            if (u[node[[r]]$j]  %in% node[[r]]$valxj) left=1
            else left=0
          }
         }

        if ( left == 1 ) {
          r <- node[[r]]$Left
        } else {
          r <- node[[r]]$Right
        }

      } # end of while

      Node[i]=r
      class[i]=node[[r]]$yhat
      if( ncaty >= 2 ) prob[i,]=node[[r]]$prob

    } # end of i loop for obs


  }# end of majestic < 0



  return( list(yhat=class, prob=prob, Node=Node) )

} # end of predict_dt



