#' Quantification of a Categorical Regressor Variable
#'
#' This function creates a set of quantified regressor variables
#' from a categorical regressor variable. \cr
#' Japanese help file: \link[lazy.tree]{quantify_x_JPH}
#'
#'
#' @param y The criterion variable
#' @param x The categorical regressor variable (as numeric)
#' @param ux The unique values of \code{x}
#' @param loc The locations of the unique values (ux) in \code{x}.
#' @param Y The dichotomized criterion matrix for \code{method=4}
#' @param method Treatment of the categorical regressor variable. \cr
#' = 0 to dummy expand the categorical regressor, \cr
#' = 1 to use the first canonical varialbe, \cr
#' = 2 to use up to the 2nd canonical variables and their sum and difference, cr
#' = 3 to use one-versus-rest dichotomization. \cr
#' = 4 to use dichotomization of the criterion variable on the basis of
#' all possible bipartitions of the levels of the criterion variable.
#' @param remove_redundant = 0 not to remove the redundant and zero variance
#' columns.
#'
#' @details
#' If the criterion is continuous or
#' the # of levels of the criterion is equal to two,
#' the conditional mean of the criterion given the level of regressor
#' will be used as the quantified regressor. \cr
#' (The # of quantified regressor = 1.)
#'
#' Regardless of the method, if the criterion is continuous or
#' the # of levels of the criterion is equal to two,
#' the conditional mean of the criterion given the level of regressor
#' will be used as the quantified regressor. \cr
#' (The # of quantified regressor = 1.)
#'
#' If \code{method=0}  and if the # of levels of the criterion
#' is greater than 2, a categorical regressor will be dummy expanded
#' and the resulting dummy variables will be used as the quantified
#' regressor variables.
#' (The # of quantified regressor = # of levels of the regressor)
#'
#' If \code{method=1}
#' When the # of levels of the criterion is greater than 2,
#' the first canonical variable of the canonical discriminant analysis
#' of the criterion on the dummy expanded regressor will be used. \cr
#' (The # of quantified regressor = 1.)
#'
#' If \code{method=2}, in addition to the above, the 2nd canonical
#' variable, if exists, and the sum and difference of those two canonical
#' variables will be added as the quantified regressors. \cr
#' (The # of quantified regressors = 4 or 1.)
#'
#' If \code{method=3}, the one-versus-rest dichotomization of the criterion
#' variable is used and the conditional means will be used as the quantified
#' regressors.
#' If any of the resulting quantified variables has zero variance
#' or duplicated,
#' the redundant quantified variables will be removed. \cr
#' (The # of quantified regressors = # of levels of the criterion
#'  or less.)
#'
#' If \code{method=4},
#' all possible bipartitions of the levels of the criterion
#' is  used for the dichotomization
#' and the conditional means will be used as the quantified regressors. \cr
#' If any of the resulting quantified variables has zero variance
#' or duplicated,
#' the redundant quantified variables will be removed. \cr
#' The columns of the quantified regressor matrices correspond to the
#' set of all bipartitions of the levels of the criterion variable. \cr
#' (The # of quantified regressors = 2^{(nc-1)}-1 or less where nc is the
#' # of levels of the criterion variable.) \cr
#' See the example below for details.
#'
#' \code{method=4} seems to be the best option when the # of levels
#'  of the criterion is not so large ( <= 10 or so. )
#'  and the # of levels of the regressor is very large. \cr
#'
#'
#'
#' @examples
#' # iris: categorical y, categorical X
#' Species <- iris$Species
#' levels(Species) <- c("set","ver","vir")
#' Xc <- iris[,-5]
#' Xc[,3] <- cut(Xc[,3],9, labels=1:9)
#' Xc[,4] <- cut(Xc[,4],9, labels=1:9)
#'
#' y <- Species
#' x <- as.numeric(Xc[,3])
#' temp <- Unique(x)
#' ux <- temp$value
#' loc <- temp$loc
#'
#' q1 <- quantify_x( y, x, ux, loc, method=1 )$newux
#' q2 <- quantify_x( y, x, ux, loc, method=2 )$newux
#'
#' Print(q1,q2, fmt="5.2")
#'
#'
#'
#' # explanatory example of the method.
#' #
#' # criterion y as factor.
#' y <- as.factor( LETTERS[c(1,2,3,4,4,4,2,2,1,1,2,3,4,3,1,3,4,2,2,2)] )
#' # regressor x as numeric even if it is categorical.
#' x <- c(1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5, 1,2,3,4,5 )
#'
#' # # of criterion categories and # of obs
#' ncaty <- length(unique(y))
#' n <- length(y)
#'
#' # unique value of x and their locations
#' temp <- Unique(x)
#' ux <- temp$value
#' loc <- temp$loc
#'
#' # method 1 and 2
#' q1 <- quantify_x( y, x, ux, loc, method=1 )$newux
#' q2 <- quantify_x( y, x, ux, loc, method=2 )$newux
#'
#' Print(ux,q1,q2, fmt="6.3")
#'
#' # convert them to rank order to check redundancy.
#' q1 <- t(unique(t(apply( round(q1,3), 2, rank, ties.method="min" ))))
#' q2 <- t(unique(t(apply( round(q2,3), 2, rank, ties.method="min" ))))
#'
#' cat("Note that q3 was dropped since its rank order is the same as q2.\n")
#' Print(ux,q1,q2, fmt="4.0")
#'
#' # Details of method=3 and method=4 quantification:
#' #
#' # In reality, Y will be generated in DecTree function, and
#' # conversion to rank will be done in Divide function.
#' #
#'
#' # 1) preparation of Y matrix: all the bipartitions of y levels
#' id <- gen01pat( ncaty, sort=-2, partition=1 ) == 1  # logical matrix
#' npart <- nrow(id)
#' uy <- unique(y)
#' Y <- matrix(0,n,npart)
#' for(k in 1:npart ){
#'  Y[,k] <- y %in% uy[id[k,]]
#' }
#' colnames(Y) <- paste("qx",1:ncol(Y), sep="")
#' cat("\nThe Base of the Dichotomization of the criterion\n")
#' printm(y,Y, x)
#'
#' # 2) conditional mean of dichotomized y according to Y matrix
#' uxnew <- matrix(0,length(ux),ncol(Y))
#' for( i in 1:length(ux) ){
#'  uxnew[i,] <- colMeans(Y[loc[[i]],,drop=0])
#' }
#' cat("\nNote that some columns may have zero variance and some columns may"
#'     , " be duplicated.\n")
#' colnames(uxnew) <- paste("qx",1:ncol(uxnew), sep="")
#' rownames(uxnew) <- paste("x",1:nrow(uxnew), sep="")
#' Print(ux, uxnew)
#'
#' # By quantify_x function (Redundant columns removed.)
#' q3 <- quantify_x( y, x, ux, loc, Y[,1:ncaty], method=3 )$newux
#' q4 <- quantify_x( y, x, ux, loc, Y, method=4 )$newux
#' cat("Note that q4 was removed.\n")
#' Print(ux,q3,q4, fmt="5.2")
#'
#' # convert them to rank order
#' q3 <- t(unique(t(apply( q3, 2, rank, ties.method="min" ))))
#' q4 <- t(unique(t(apply( q4, 2, rank, ties.method="min" ))))
#' cat("In this case, all the rank orders are different.\n")
#' Print(ux,q3,q4, fmt="4.0")
#'
#'
#' @return
#' A list of:
#' \preformatted{
#' \code{ncatx} # of levels of the regressor
#' \code{ncaty} # of levels of the criterion
#' \code{newux} Quantified regressor variable in a matrix with
#'  dyplicated and zero-variance columns removed.
#' \code{candisc} = -2 if \code{method=4} is used.
#' = -1 if no quantification is done.
#' = 0 if the conditional mean of the criterion is used.
#' = 1 if the canonical variables is used.
#' }
#'
#' @export
#'

quantify_x <- function( y, x, ux, loc, Y, method=1, remove_redundant=1 ){
  # quantify x if necessary
  # Shin-ichi Mayekawa
  # 20230905,06
  # as.numeric(y), yn: 20230909
  # replace eignum by method: 20230911
  # method=2: 2023091
  # do not renove the first col of Xd: 20230914
  # method=2,3: 20230914,15
  # return candisc flag: 20230928
  # Y matrix: 20231005
  # NULL from candisc: 20231007cot
  # candisc eve and model definition: 20231009
  # method=4 when ncaty<=2: 20231018
  # new method=3: 20231108
  # roxygen2: 20231109,1112cot
  # remove_redundant, col/row names: 20231112cot
  # roxygen2: 20231120
  # reorganized: 20231218
  # method=0: 20231218,19
  # lost modification of 20231219? applied: 20240317
  #
  #

  # Args:
  #  (y, x)  pair of original variables
  #  ux      the unique values of x
  #  loc     the original locations of the unique values
  #  method  = 1 to return a vector
  #          = 2 to return a matrix
  #
  # variable type by # of categories:
  #    ncat =  0 or 1  continuous
  #    ncat =  2 binary
  #    ncat >= 3 polytomous
  #
  #
  # method = 1  The first candisc var
  #        = 2  1st, 2nd and 1st+2nd and 1st-2nd candisc var
  #        = 3  dichotomization ov y (one and the rest)
  #        = 4  all possible dichotomization of y
  #

  #
  # x is assumed to be numeric, not factor
  #

  # check
  if( !(method %in% 0:4) ){
   cat("\nerror1:(quantify_x) method must be 0, 1, 2, 3, or 4.\n")
   stop()
  }

  # # of categories
  ncatx=length(ux)
  # lenux=length(ux)
  ncaty=length(unique(y))



  # if x is continuous or binary, do nothing.
  if( ncatx <= 2 ){
    res=list( ncatx=ncatx, ncaty=ncaty, newux=matrix(ux,,1), candisc=-1 )
    return( res )
  } # end of ncatx <= 2


  # Below, ncatx >= 3, polytomous regressor.

  # all methods
  if( ncaty <= 2 ){

    # When  ncaty <= 2, regardless of the method,
    # calculate the conditional mean of y given the levels of x.

    candisc=0

    # numeric values of a factor
    if( is.factor(y) ) yn=unclass(y)
    else yn=y

    # conditional mean of y
    uxnew=matrix(0,ncatx,1)
    for( i in 1:ncatx ){
      uxnew[i,1]=mean( yn[loc[[i]]] )
    }
    colnames(uxnew)="qx1"
    res=list( ncatx=ncatx, ncaty=ncaty, newux=uxnew, candisc=0 )
    return( res )

  } # end of ncaty <= 2


  # Here, ncaty >= 3  and  ncatx >= 3, polytomous regressor.

  # method = 0:  dummy-expand
  if( method == 0 ){

    uxnew=design_expand( ux, type=-1)
    colnames(uxnew)=paste("qx",1:ncol(uxnew), sep="")
    res=list( ncatx=ncatx, ncaty=ncaty, newux=uxnew, candisc=0 )
    return( res )

  } # end of method=0


  # method = 3 and 4: use Y
  # Depending on the method, DecTree function creates
  # different Y matrix.
  #
  if( ncaty >= 3  &&  method >= 3 ){

    uxnew=matrix(0,ncatx,ncol(Y))
    for( i in 1:ncatx ){
      uxnew[i,]=colMeans(Y[loc[[i]],,drop=0])
    }

    colnames(uxnew)=paste("qx",1:ncol(uxnew), sep="")
    rownames(uxnew)=paste("x",1:nrow(uxnew), sep="")

    if( remove_redundant == 1 ){
     locr=which(apply(uxnew,2,var) > 0)
     uxnew=uxnew[,locr, drop=0]
     locr=!duplicated(t(uxnew))
     uxnew=uxnew[,locr, drop=0]
     # ncol(uxnew) can be zero!!   20231219
    }

    res=list( ncatx=ncatx, ncaty=ncaty, newux=uxnew, candisc=-2 )
    return( res )

  } # end of method=3 and 4



  if( method %in% c(1,2) ){
    # here, ncaty >= 3
    # method = 1, 2 (3 or 4 were already done.)
    # candisc
    #

    candisc=1

    if( method >= 1 ){

      # one candisc
      # dummy expand xsj
      factx=factor(x, levels = min(x):max(x))

      Xd=model.matrix( ~factx )[,, drop=0][,-1]
      eve=candisc( y, Xd, print=0 )$eve

      if( !is.null(eve) ){
       ndim=ncol(eve)
       factx=factor(ux, levels = min(ux):max(ux))
       uXd=model.matrix( ~ factx - 1)[,, drop=0][,-1]
       canvar=uXd%*%eve[,1:min(2,ndim),drop=0]
      }
      else{
       ndim=0
       canvar=matrix(ux,,1)
      }

      uxnew=canvar[,1,drop=0]

      if( method >= 2 && ndim >= 2 ){
        uxnew=cbind( uxnew, canvar[,2]
                    , canvar[,1]+canvar[,2], canvar[,1]-canvar[,2] )
      }

    } # end of method=1 or 2


  } # end of method=1 or 2

  colnames(uxnew)=paste("qx",1:ncol(uxnew), sep="")

  res=list( ncatx=ncatx, ncaty=ncaty, newux=uxnew, candisc=candisc )
  return( res )


} # end of quantify_x



