#' Similarity Index of Two Trees
#'
#' Returns the maximum depth at which the same splitting variables are used.
#' Japanese help file: \link[lazy.tree]{treesim_dt_JPH}
#'
#' @param T1 first class DecTree object
#' @param T2 second class DecTree object
#' @param print = 1 to print the result
#'
#'
#' @return A list of
#' \preformatted{
#' simdepth The maximum depth at which the same splitting variables are used.
#' Eq  A vector indicating if all the splitting variables are the same.
#' Eqp A vector indicating if all the splitting variables are the same.
#' EqTable A depth by node matrix indicating if the splitting variables
#'  are the same.
#' }
#'
#' @export
#'

treesim_dt <- function( T1, T2, print=0 ){
 # tree similarity
 # Shin-ichi Mayekawa
 # 20240307
 # Eq and print: 20240407cot
 #

 maxdepth=min(T1$MaxDepth,T2$MaxDepth)
 EqTable=matrix(NA, maxdepth-1,2^(maxdepth-1))
 Eqp=numeric(maxdepth-1)
 Eq=numeric(maxdepth-1)
 simdepth=maxdepth
 for( dp in 1:(maxdepth-1) ){

  i1=node_info_dt( T1, depth=dp, nohist=1, print=0 )
  i2=node_info_dt( T2, depth=dp, nohist=1, print=0 )

  if( simdepth == maxdepth  &&  length(i1) != length(i2) ){
   simdepth=dp
  }

  eq=i1[,"SplitVar"] == i2[,"SplitVar"]
  Eq[dp]=sum(eq)
  eqp=sum(eq)/length(eq)
  Eqp[dp]=eqp

  EqTable[dp,1:length(eq)]=eq

  rownames(EqTable)=paste("d",1:(maxdepth-1),sep="")
  colnames(EqTable)=paste("n",1:2^(maxdepth-1),sep="")
  names(Eq)=paste("d",1:(maxdepth-1),sep="")
  names(Eqp)=paste("d",1:(maxdepth-1),sep="")

  if( simdepth == maxdepth  &&  eqp < 1 ){
   simdepth=dp
  }

 } # end of dp loop

 simdepth=simdepth-1

 if( print > 0 ){
  Print(simdepth, Eq, Eqp)
  Print(EqTable)
 }

 res=named_list( simdepth, Eq, Eqp, EqTable )

 return( res )

} # end of treesim_dt


